import os
from hashlib import sha256

from PIL import Image, ExifTags
from io import BytesIO

from dropbox import DropboxOAuth2FlowNoRedirect
from dropbox.dropbox_client import Dropbox
from dropbox.files import FolderMetadata
from dropbox.files import DownloadError

import ffmpeg
from pymemri.pod.client import PodClient

from dropbox_plugin.data.schema import Photo, Video, Folder, File


IMAGE_FORMATS = ('jpeg', 'tiff', 'jpg', 'png', 'bmp', 'ico', 'img', )
VIDEO_FORMATS = ('webm', 'mkv', 'flv', 'avi', 'wmv', 'mp4', )


class DropboxService:

    def __init__(self, client: PodClient):
        self.client = client
        self.access_token = None
        self.dbx = None
        self.auth_flow = None

        self.media_types = {
            IMAGE_FORMATS: {'get_info_func': self.get_image_data, 'item_type': Photo},
            VIDEO_FORMATS: {'get_info_func': self.get_video_data, 'item_type': Video},
        }

    def get_image_data(self, meta, response):
        image = Image.open(BytesIO(response.content))
        if bool(image._getexif()):
            exif = {
                ExifTags.TAGS[k]: v
                for k, v in image._getexif().items()
                if k in ExifTags.TAGS
            }
        else:
            exif = {}
        attrs_dict = {
            'fileLocation': meta.path_display,
            'exifData': str(exif),
            'name': meta.name,
            'height': image.height,
            'width': image.width,
            'channels': len(image.getbands()),
        }
        return image.tobytes(), attrs_dict

    def get_video_data(self, meta, response):
        filename = meta.name
        with open(filename, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024):
                fd.write(chunk)
        probe = ffmpeg.probe(filename)
        video_bytes = open(filename, 'rb').read()
        attrs_dict = {
            'bitrate': int(probe['format']['bit_rate']),
            'duration': int(probe['format']['duration'].split('.')[0]),
            'fileLocation': meta.path_display,
            'name': meta.name,
        }
        os.remove(filename)
        return video_bytes, attrs_dict

    def get_file_data(self, meta, response):
        attrs_dict = {
            'fileLocation': meta.path_display,
            'filename': meta.name,
        }
        return response.content, attrs_dict

    def create_file(self, file_bytes):
        sha = sha256(file_bytes).hexdigest()
        file = File.from_data(sha256=sha)
        self.client.create(file)
        if not self.client.get_file(sha):
            self.client.upload_file(file_bytes)
        return file

    def check_item_duplicates(self, item):
        existing_items = self.client.search({'type': type(item).__name__})
        for existing_item in existing_items:
            if len(existing_item.file) > 0:
                if item.fileLocation == existing_item.fileLocation and item.file[0].sha256 == existing_item.file[0].sha256:
                    return True
        return False

    def _create_media_item(self, item_type, file_bytes, item_attrs, folder=None):
        item = item_type.from_data(**item_attrs)
        if folder:
            item.add_edge('folder', folder)
        file = self.create_file(file_bytes)
        item.add_edge('file', file)
        created = False
        if not self.check_item_duplicates(item):
            self.client.create(item)
            created = True
        return created, item

    def check_file_duplicates(self, file):
        existing_items = self.client.search({'type': type(file).__name__})
        for existing_item in existing_items:
            if file.fileLocation == existing_item.fileLocation and file.sha256 == existing_item.sha256:
                return True
        return False

    def _create_file_instance(self, media_bytes, attrs_dict, folder):
        sha = sha256(media_bytes).hexdigest()
        attrs_dict['sha256'] = sha
        file = File.from_data(**attrs_dict)
        if folder:
            file.add_edge('folder', folder)
        created = False
        if not self.check_file_duplicates(file):
            self.client.create(file)
            created = True
        if not self.client.get_file(sha):
            self.client.upload_file(media_bytes)

        return created, file

    def get_item_media_type_helper(self, meta):
        item_extension = meta.name.split('.')[-1]
        for key in self.media_types.keys():
            if item_extension in key:
                return self.media_types[key]
        return {'get_info_func': self.get_file_data, 'item_type': None}

    def create_media_item(self, item_meta, folder):
        meta, response = self.dbx.files_download(item_meta.path_display)
        type_helper = self.get_item_media_type_helper(meta)
        media_bytes, attrs_dict = type_helper['get_info_func'](meta, response)
        if type_helper['item_type']:
            created, item = self._create_media_item(type_helper['item_type'], media_bytes, attrs_dict, folder)
        else:
            created, item = self._create_file_instance(media_bytes, attrs_dict, folder)

        if not created:
            print(f'{meta.name} not been created')
        return item

    def create_folder(self, name, path, parent_folder=None):
        existing_items = self.client.search({'type': 'Folder'})
        for existing_item in existing_items:
            if existing_item.path == path:
                return existing_item
        folder = Folder.from_data(name=name, path=path)
        if parent_folder:
            folder.add_edge('folder', parent_folder)
        self.client.create(folder)
        return folder

    def save_folder_content(self, directory, parent_folder=None):
        if self.dbx is None:
            print('Not logged in')
            return
        content = self.dbx.files_list_folder(directory, recursive=False)
        while True:
            for entry in content.entries:
                if isinstance(entry, FolderMetadata):
                    folder = self.create_folder(entry.name, entry.path_display, parent_folder)
                    self.save_folder_content(entry.path_display, folder)
                else:
                    try:
                        self.create_media_item(entry, parent_folder)
                        print(f'Saved {entry.path_display}')
                    except Exception as e:
                        print(e)
                        print(f'Error while downloading {entry.path_display}')
            if not content.has_more:
                break
            content = self.dbx.files_list_folder_continue(content.cursor)

    def get_oauth_url(self, app_id, app_secret):
        self.auth_flow = DropboxOAuth2FlowNoRedirect(app_id, app_secret)
        authorize_url = self.auth_flow.start()
        return authorize_url

    def get_oauth_token(self, auth_code):
        try:
            oauth_result = self.auth_flow.finish(auth_code)
        except Exception as e:
            print(e)
            return {'success': False}
        self.access_token = oauth_result.access_token
        self.dbx = Dropbox(self.access_token)
        return {'refreshToken': oauth_result.refresh_token, 'token': oauth_result.access_token, 'success': True}

